# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides the PyAnsys Geometry ``geometry`` subpackage."""

from ansys.geometry.core.shapes.curves.circle import Circle, CircleEvaluation
from ansys.geometry.core.shapes.curves.curve import Curve
from ansys.geometry.core.shapes.curves.ellipse import Ellipse, EllipseEvaluation
from ansys.geometry.core.shapes.curves.line import Line, LineEvaluation
from ansys.geometry.core.shapes.curves.nurbs import NURBSCurve, NURBSCurveEvaluation
from ansys.geometry.core.shapes.parameterization import (
    Interval,
    Parameterization,
    ParamForm,
    ParamType,
    ParamUV,
)
from ansys.geometry.core.shapes.surfaces.cone import Cone, ConeEvaluation
from ansys.geometry.core.shapes.surfaces.cylinder import Cylinder, CylinderEvaluation
from ansys.geometry.core.shapes.surfaces.plane import PlaneEvaluation, PlaneSurface
from ansys.geometry.core.shapes.surfaces.sphere import Sphere, SphereEvaluation
from ansys.geometry.core.shapes.surfaces.surface import Surface
from ansys.geometry.core.shapes.surfaces.surface_evaluation import SurfaceEvaluation
from ansys.geometry.core.shapes.surfaces.torus import Torus
from ansys.geometry.core.shapes.surfaces.trimmed_surface import TrimmedSurface
